@@ -1,7 +1,7 @@
 /* bsect.c  -  Boot sector handling */
 /*
 Copyright 1992-1998 Werner Almesberger.
-Copyright 1999-2004 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -100,17 +100,16 @@ BUILTIN_FILE *select_loader(void)
     else if (strstr(install,"bmp") || bitmap) loader = &Bitmap;   /* menu interface (MDA, EGA, VGA) */
 	
     adapter = get_video();	/* from probe.c */
-    if (!nowarn) {
-	if (adapter <= VIDEO_UNKNOWN && (verbose>=2 || loader==&Bitmap))
-	    fprintf(errstd, "Warning: Unable to determine video adapter in use in the present system.\n");
-	if (loader == &Third && adapter == VIDEO_CGA)
-	    fprintf(errstd, "Warning: Video adapter (CGA) is incompatible with the boot loader selected for\n"
-	    			"  installation ('install = menu').\n");
-	if (loader == &Bitmap && adapter < VIDEO_MCGA && adapter > VIDEO_UNKNOWN)
-	    fprintf(errstd, "Warning: Video adapter (%s) is incompatible with the boot loader selected for\n"
-	    			"  installation ('install = bitmap').\n", adapter==VIDEO_CGA ? "CGA" :
+
+    if (adapter <= VIDEO_UNKNOWN && (verbose>=2 || loader==&Bitmap))
+	    warn("Unable to determine video adapter in use in the present system.");
+    if (loader == &Third && adapter == VIDEO_CGA)
+	    warn("Video adapter (CGA) is incompatible with the boot loader selected for\n"
+	    			"  installation ('install = menu').");
+    if (loader == &Bitmap && adapter < VIDEO_MCGA && adapter > VIDEO_UNKNOWN)
+	    warn("Video adapter (%s) is incompatible with the boot loader selected for\n"
+	    			"  installation ('install = bitmap').", adapter==VIDEO_CGA ? "CGA" :
 	    			adapter==VIDEO_MDA ? "MDA" : "unknown");
-    }
     
     return loader;
 }
@@ -131,6 +130,9 @@ static char *disallow[] = {
     int error = 0;
 
     if (verbose >= 5) printf("check_options: \"%s\"\n", options);
+    if (strlen(options) > COMMAND_LINE_SIZE-1) {
+	warn("Command line options > %d will be truncated.", COMMAND_LINE_SIZE-1);
+    }
     while (*dis) {
 	here = strstr(options, *dis);
 	if (here  &&  (here==options || here[-1] == ' ') ) {
@@ -146,8 +148,8 @@ static char *disallow[] = {
 	}
 	if (error>1) die ("APPEND or LITERAL may not contain \"%s\"", *dis);
 #if !MEM_OKAY
-	if (error==1 && !nowarn) {
-	    fprintf(errstd, "Warning: APPEND or LITERAL may not contain \"mem=\"\n");
+	if (error==1) {
+	    warn("APPEND or LITERAL may not contain \"mem=\".");
 	    error = 0;
 	}
 #endif
@@ -235,9 +237,8 @@ void bmp_do_table(char *cp, MENUTABLE *menu)
     menu->mincol = getval(&cp, 1, menu->maxcol, menu->mincol, 0);
 #endif
     if ((menu->row + menu->maxcol*16 > 480 || 
-	 menu->col + (MAX_IMAGE_NAME+1)*8 + (menu->ncol-1)*menu->xpitch > 640)
-	  && !nowarn)  
-       fprintf(errstd,"Warning: 'bmp-table' may spill off screen\n");
+	 menu->col + (MAX_IMAGE_NAME+1)*8 + (menu->ncol-1)*menu->xpitch > 640))  
+	         warn("'bmp-table' may spill off screen");
 }
 
 
@@ -324,13 +325,13 @@ static void hash_password(char *password, int crcval[])
 	
 #ifdef SHS_PASSWORDS
 	shsInit();
-	shsUpdate(password, i);
+	shsUpdate((BYTE*)password, i);
 	shsFinal();
 #endif
 	for (j=0; j<MAX_PW_CRC; j++) {
 	    crcval[j] = crc =
 #ifdef CRC_PASSWORDS
-	    			crc32(password, i, poly[j]);
+	    			crc32((unsigned char *)password, i, poly[j]);
   #define PWTYPE "CRC-32"
 #else
 				shsInfo.digest[j];
@@ -460,12 +461,8 @@ static void open_bsect(char *boot_dev)
     boot_devnam = boot_dev;
     if (!boot_dev || !strcmp(boot_dev, "/") ) {
 	if (stat("/",&st) < 0) pdie("stat /");
-#if 0
-	if ((st.st_dev & PART_MASK) > PART_MAX)
-#else
 	if (MAJOR(st.st_dev) != MAJOR_MD &&
 		(st.st_dev & P_MASK(st.st_dev)) > PART_MAX)
-#endif
 	    die("Can't put the boot sector on logical partition 0x%04X",
 	      (int)st.st_dev);
 	fd = dev_open(&dev,boot_dev_nr = st.st_dev,O_RDWR);
@@ -489,8 +486,8 @@ static void open_bsect(char *boot_dev)
     geo_close(&geo);
 }
     
-    if (boot_dev_nr && !is_first(boot_dev_nr) && !nowarn)
-	fprintf(errstd,"Warning: %s is not on the first disk\n",boot_dev ?
+    if (boot_dev_nr && !is_first(boot_dev_nr) )
+	warn("%s is not on the first disk",boot_dev ?
 	  boot_dev : "current root");
     if (read(fd,(char *) &bsect,SECTOR_SIZE) != SECTOR_SIZE)
 	die("read %s: %s",boot_dev ? boot_dev : dev.name,strerror(errno));
@@ -499,12 +496,13 @@ static void open_bsect(char *boot_dev)
     if (ireloc == PTW_DOS + PTW_NTFS)  ireloc = PTW_DOS;
 /* check for override (-F) command line flag */
     if (ireloc>PTW_DOS && force_fs) {
-#if 0
-	fprintf(errstd, "WARNING: '-F' override used. Filesystem on  %s  may be destroyed.\n", boot_devnam);
-#else
-	fprintf(stderr, "WARNING: '-F' override used. Filesystem on  %s  may be destroyed.\n", boot_devnam);
+        int nowarn2 = nowarn;
+        nowarn = 0;
+
+	warn("'-F' override used. Filesystem on  %s  may be destroyed.", boot_devnam);
 	if (!yesno("\nProceed? ",0)) exit(0);
-#endif
+
+        nowarn = nowarn2;
 	ireloc=PTW_OKAY;
     }
 }
@@ -542,15 +540,13 @@ static void menu_do_scheme(char *scheme, MENUTABLE *menu)
     	    if (ispunct(*scheme)) scheme++;
     	    else die("Invalid menu-scheme punctuation");
     	}
-    	if (bg>=8 && !nowarn)
-    	    fprintf(errstd, "Warning: menu-scheme BG color may not be intensified\n");
+    	if (bg>=8)
+    	    warn("menu-scheme BG color may not be intensified");
     	*at++ = ((bg<<4) | fg) & 0x7F;
     }
     /* check on the TEXT color */
     if (menu->at_text == 0) {
-	if (!nowarn)
-	    fprintf(errstd, "Warning: menu-scheme \"black on black\" changed to "
-		"\"white on black\"\n");
+        warn("menu-scheme \"black on black\" changed to \"white on black\"");
 	menu->at_text = 0x07;
     }
     /* check on the HIGHLIGHT color */
@@ -609,7 +605,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
     	die("Filesystem would be destroyed by LILO boot sector: %s", boot_dev);
     }
     else if (ireloc==PTW_DOS) {
-	if (!nowarn) fprintf(errstd,"Warning: boot record relocation beyond BPB is necessary: %s\n", boot_dev);
+	warn("boot record relocation beyond BPB is necessary: %s", boot_dev);
     }
 
 #ifdef EARLY_MAP
@@ -659,7 +655,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 	memcpy(secondary_map+SECTOR_SIZE-4, EX_MAG_STRING, 4);
 
 	/* fill in full size of secondary boot loader in paragraphs */
-	/*bsect.par_1.*/dataend = (sectors + 5 + MAX_DESCR_SECS) * (SECTOR_SIZE/16);
+	/*bsect.par_1.*/dataend = (sectors + 5 + (COMMAND_LINE_SIZE>256) + MAX_DESCR_SECS) * (SECTOR_SIZE/16);
 
 	if (verbose > 1)
 	    printf("Secondary loader: %d sector%s (0x%0X dataend).\n",sectors,sectors == 1 ?
@@ -700,8 +696,8 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 		 || (do_md_install && !(extra==X_MBR_ONLY ))
 		  ) )
 	    bsect.par_1.prompt |= FLAG_MAP_ON_BOOT;
-	if (bios_boot!=bios_map && !nowarn)
-	    fprintf(errstd, "Warning: The boot sector and map file are on different disks.\n");
+	if (bios_boot!=bios_map)
+	    warn("The boot sector and map file are on different disks.");
 	if ( (bios_map & 0x80) && !do_md_install &&
 		!cfg_get_flag(cf_options, "static-bios-codes") ) /* hard disk & not raid master */
 	    bsect.par_1.map_serial_no = serial_no[bios_map - 0x80];
@@ -725,7 +721,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
     if (bitmap) {
 	message = scheme;
 	if (!(stage_flags & STAGE_FLAG_BMP4)) {
-	    if (!nowarn) fprintf(errstd,"Warning: Non-bitmap capable boot loader; 'bitmap=' ignored.\n");
+	    warn("Non-bitmap capable boot loader; 'bitmap=' ignored.");
 	    message = NULL;
 	}
     }
@@ -759,9 +755,9 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 		    	((bits = bmh->numBitPlanes * bmh->numBitsPerPlane) == 4 ||
 		    	  bits == 8) ) {
 		    if (!bitmap) die("Message specifies a bitmap file");
-		    if (bits>4 && adapter<VIDEO_VESA && !nowarn)
-			fprintf(errstd,"Warning: Video adapter does not support VESA BIOS extensions needed for\n"
-				       "  display of 256 colors.  Boot loader will fall back to TEXT only operation.\n");
+		    if (bits>4 && adapter<VIDEO_VESA)
+			warn("Video adapter does not support VESA BIOS extensions needed for\n"
+				       "  display of 256 colors.  Boot loader will fall back to TEXT only operation.");
 		}
 		else if (bitmap) die("Unsupported bitmap");
 	    } else if (bitmap) die("Not a bitmap file");
@@ -788,7 +784,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
     if (cfg_get_flag(cf_options,"unattended")) {
 	param2.flag2 |= FLAG2_UNATTENDED;
 	if (timeout < 0) {
-	    fprintf(errstd,"Warning: UNATTENDED used; setting TIMEOUT to 200 (20 seconds)\n");
+	    warn("UNATTENDED used; setting TIMEOUT to 20s (seconds).");
 	    timeout = 200;
 	}
     }
@@ -837,13 +833,13 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 						(int)param2.ser_param);
 	}
 	if (delay < 20 && !cfg_get_flag(cf_options,"prompt")) {
-	    fprintf(errstd,"Warning: no PROMPT with SERIAL; setting DELAY to 20 (2 seconds)\n");
+	    warn("no PROMPT with SERIAL; setting DELAY to 20 (2 seconds)");
 	    delay = 20;
 	}
     }
     bsect.par_1.prompt |= cfg_get_flag(cf_options,"prompt") ? FLAG_PROMPT : 0;
     if (cfg_get_flag(cf_options,"suppress-boot-time-BIOS-data")) {
-    	fprintf(errstd,"CAUTION:  boot-time BIOS data will not be saved\n");
+    	warn("boot-time BIOS data will not be saved.");
     	bsect.par_1.prompt |= FLAG_NOBD;
     }
     if (!fetch() && (bios_map &
@@ -861,7 +857,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 #ifndef LCF_INITRDLOW
 	bsect.par_1.prompt |= FLAG_LARGEMEM;
 #else
-	if (!nowarn) fprintf(errstd, "Warning: This LILO compiled with INITRDLOW option, 'large-memory' ignored\n");
+	warn("This LILO compiled with INITRDLOW option, 'large-memory' ignored.");
 #endif
     }
     bsect.par_1.prompt |= raid_flags;
@@ -905,21 +901,21 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
     memcpy(&(menu->row), &(lhv.row), sizeof(lhv) - sizeof(lhv.size) - sizeof(lhv.magic));
 
     if ((scheme = cfg_get_strg(cf_options,"menu-scheme"))) {
-	if (!(stage_flags & STAGE_FLAG_MENU) && !nowarn)
-	    fprintf(errstd,"Warning: 'menu-scheme' not supported by boot loader\n");
+	if (!(stage_flags & STAGE_FLAG_MENU))
+	    warn("'menu-scheme' not supported by boot loader");
     	menu_do_scheme(scheme, menu);
     }
     if ((scheme = cfg_get_strg(cf_options,"menu-title"))) {
-	if (!(stage_flags & STAGE_FLAG_MENU) && !nowarn)
-	    fprintf(errstd,"Warning: 'menu-title' not supported by boot loader\n");
-	if (strlen(scheme) > MAX_MENU_TITLE && !nowarn)
-	    fprintf(errstd,"Warning: menu-title is > %d characters\n", MAX_MENU_TITLE);
+	if (!(stage_flags & STAGE_FLAG_MENU))
+	    warn("'menu-title' not supported by boot loader");
+	if (strlen(scheme) > MAX_MENU_TITLE)
+	    warn("menu-title is > %d characters", MAX_MENU_TITLE);
     	strncpy(menu->title, scheme, MAX_MENU_TITLE);
     	menu->len_title = strlen(menu->title);
     }
     if ((scheme = cfg_get_strg(cf_options,"bmp-table"))) {
-	if (!(stage_flags & STAGE_FLAG_BMP4) && !nowarn)
-	    fprintf(errstd,"Warning: 'bmp-table' not supported by boot loader\n");
+	if (!(stage_flags & STAGE_FLAG_BMP4))
+	    warn("'bmp-table' not supported by boot loader");
     }
     bmp_do_table(scheme, menu);
     if (bitmap) {
@@ -927,13 +923,13 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 	if (verbose>=3) printf("image_menu_space = %d\n", image_menu_space);
     }
     if ((scheme = cfg_get_strg(cf_options,"bmp-colors"))) {
-	if (!(stage_flags & STAGE_FLAG_BMP4) && !nowarn)
-	    fprintf(errstd,"Warning: 'bmp-colors' not supported by boot loader\n");
+	if (!(stage_flags & STAGE_FLAG_BMP4))
+	    warn("'bmp-colors' not supported by boot loader");
     }
     bmp_do_colors(scheme, menu);
     if ((scheme = cfg_get_strg(cf_options,"bmp-timer"))) {
-	if (!(stage_flags & STAGE_FLAG_BMP4) && !nowarn)
-	    fprintf(errstd,"Warning: 'bmp-timer' not supported by boot loader\n");
+	if (!(stage_flags & STAGE_FLAG_BMP4))
+	    warn("'bmp-timer' not supported by boot loader");
     }
     bmp_do_timer(scheme, menu);
 #if 0
@@ -971,7 +967,7 @@ static int get_image(char *name,char *label,IMAGE_DESCR *descr)
     }
     if (strchr(label,' ')) die("Image name, label, or alias contains a blank character: '%s'", label);
     if (strlen(label) > MAX_IMAGE_NAME) die("Image name, label, or alias is too long: '%s'",label);
-    for (uch=label; *uch; uch++) {
+    for (uch=(unsigned char*)label; *uch; uch++) {
 	if (*uch<' ')  die("Image name, label, or alias contains an illegal character: '%s'", label);
     }
     for (other = image_base; other <= image; other++) {
@@ -1019,6 +1015,18 @@ static int get_image(char *name,char *label,IMAGE_DESCR *descr)
 		!strcmp(deflt,label))
 #endif
 	descrs.d.descr[this_image].flags |= FLAG_VMDEFAULT;
+	param2.flag2 |= FLAG2_VIRTUAL;
+#endif
+
+#ifdef LCF_NOKEYBOARD
+    if ( (deflt = cfg_get_strg(cf_options,"nokbdefault")) &&
+#ifdef LCF_IGNORECASE
+		!strcasecmp(deflt,label))
+#else
+		!strcmp(deflt,label))
+#endif
+	descrs.d.descr[this_image].flags |= FLAG_NOKBDEFAULT;
+	param2.flag2 |= FLAG2_NOKBD;
 #endif
 
     return this_image;
@@ -1071,8 +1079,7 @@ if (image) { /* long section specific to 'image=' */
     if ((vga = cfg_get_strg(cf_kernel,"vga")) || (vga = cfg_get_strg(cf_options,
       "vga"))) {
 #ifndef NORMAL_VGA
-	if (!nowarn)
-	    fprintf(errstd,"Warning: VGA mode presetting is not supported; ignoring 'vga='\n");
+	warn("VGA mode presetting is not supported; ignoring 'vga='");
 #else
 	descr->flags |= FLAG_VGA;
 	     if (!strcasecmp(vga,"normal")) descr->vga_mode = NORMAL_VGA;
@@ -1092,14 +1099,17 @@ if (image) { /* long section specific to 'image=' */
 #endif
     append_local = cfg_get_strg(cf_options,"append");	/* global, actually */
     if ((append = cfg_get_strg(cf_kernel,"append")) ||
-	(append = append_local)  )
-				strcat(strcat(options,append)," ");
+	(append = append_local)  ) {
+		if (strlen(append) > COMMAND_LINE_SIZE-1) die("Command line options > %d", COMMAND_LINE_SIZE-1);
+		strcat(strcat(options,append)," ");
+    }
 
 #if 1
     append = append_local;	/* append == global append */
     if ((append_local = cfg_get_strg(cf_kernel,"addappend"))) {
-	if (!append && !nowarn)
-	    fprintf(errstd, "Warning: ADDAPPEND used without global APPEND\n");
+	if (!append)
+	    warn("ADDAPPEND used without global APPEND");
+	if (strlen(options)+strlen(append_local) > SECTOR_SIZE-1) die("Command line options > %d", COMMAND_LINE_SIZE-1);
 	strcat(options,append_local);
     }
 #endif
@@ -1150,11 +1160,23 @@ if (image) { /* long section specific to 'image=' */
     }
 
 #ifdef LCF_VIRTUAL
-    if (cfg_get_flag(cf_all,"vmwarn")) descr->flags |= FLAG_VMWARN;
-    if (cfg_get_flag(cf_all,"vmdisable")) descr->flags |= FLAG_VMDISABLE;
+    if (cfg_get_flag(cf_all,"vmwarn")) {
+        descr->flags |= FLAG_VMWARN;
+	param2.flag2 |= FLAG2_VIRTUAL;
+    }
+    if (cfg_get_flag(cf_all,"vmdisable")) {
+        descr->flags |= FLAG_VMDISABLE;
+	param2.flag2 |= FLAG2_VIRTUAL;
+    }
     if ( (descr->flags & FLAG_VMWARN) && (descr->flags & FLAG_VMDISABLE) )
 	die ("VMWARN and VMDISABLE are not valid together");
 #endif
+#ifdef LCF_NOKEYBOARD
+    if (cfg_get_flag(cf_all,"nokbdisable")) {
+        descr->flags |= FLAG_NOKBDISABLE;
+	param2.flag2 |= FLAG2_NOKBD;
+    }
+#endif
 
 #if 1
     if (cfg_get_flag(cf_all,"mandatory") || cfg_get_flag(cf_options,
@@ -1168,8 +1190,8 @@ if (image) { /* long section specific to 'image=' */
 	    descr->flags |= FLAG_RESTR;
     }
     if (password && *password && config_read) {
-	fprintf(errstd,"Warning: %s should be readable only "
-	  "for root if using PASSWORD\n",config_file);
+	warn("%s should be readable only "
+	  "for root if using PASSWORD", config_file);
 	config_read = 0;	/* suppress further warnings */
     }
 #else
@@ -1234,6 +1256,11 @@ static void bsect_done(char *name,IMAGE_DESCR *descr)
 	if (descrs.d.descr[this_image].flags & FLAG_VMDEFAULT ||
 		(this>=0 && (descrs.d.descr[this].flags & FLAG_VMDEFAULT)) )
 	    printf(" @");
+#endif
+#ifdef LCF_NOKEYBOARD
+	if (descrs.d.descr[this_image].flags & FLAG_NOKBDEFAULT ||
+		(this>=0 && (descrs.d.descr[this].flags & FLAG_NOKBDEFAULT)) )
+	    printf(" &");
 #endif
 	if (descrs.d.descr[this_image].flags & FLAG_TOOBIG ||
 		(this>=0 && (descrs.d.descr[this].flags & FLAG_TOOBIG)) )
@@ -1262,12 +1289,58 @@ int bsect_number(void)
 
 static void unbootable(void)
 {
+#if 0
     fflush(stdout);
-    fprintf(errstd,"\nWARNING: The system is unbootable !\n");
+    fprintf(errstd,"\nCAUTION: The system is unbootable !\n");
     fprintf(errstd,"%9sRun LILO again to correct this.","");
+#else
+    warn("The system is unbootable !\n"
+          "\t Run LILO again to correct this.");
+#endif    
 }
 
 
+#ifdef LCF_VIRTUAL
+void check_vmdefault(void)
+{
+    char * deflt;
+    int i;
+    
+    if ( (deflt = cfg_get_strg(cf_options,"vmdefault")) ) {
+	for (i=0; i<image; ++i) {
+	    if (descrs.d.descr[i].flags & FLAG_VMDEFAULT) {
+	        if (descrs.d.descr[i].flags & FLAG_VMDISABLE)
+		    die("VMDEFAULT image cannot have VMDISABLE flag set");
+
+                return;
+	    }
+	}
+	die("VMDEFAULT image does not exist.");
+    }
+}
+#endif
+
+#ifdef LCF_NOKEYBOARD
+void check_nokbdefault(void)
+{
+    char * deflt;
+    int i;
+    
+    if ( (deflt = cfg_get_strg(cf_options,"nokbdefault")) ) {
+	for (i=0; i<image; ++i) {
+	    if (descrs.d.descr[i].flags & FLAG_NOKBDEFAULT) {
+		if (descrs.d.descr[i].flags & FLAG_NOKBDISABLE)
+		    die("NOKBDEFAULT image cannot have NOKBDISABLE flag set");
+
+		return;
+	    }
+	}
+	die("NOKBDEFAULT image does not exist.");
+    }
+}
+#endif
+
+
 void check_fallback(void)
 {
     char *start,*end;
@@ -1317,7 +1390,7 @@ static int timestamp = 0;
     if (pass<1) {	/* BIOS_TT logic */
 	MENUTABLE *menu = &menuparams;
 	map_descrs(&descrs, menu->mt_descr, &menuparams.dflcmd);
-	menuparams.raid_dev_mask = raid_mask(menuparams.raid_offset);
+	menuparams.raid_dev_mask = raid_mask((int*)menuparams.raid_offset);
 	memcpy(menuparams.serial_no, serial_no, sizeof(serial_no));
 	memcpy(table+256, &menuparams, sizeof(menuparams));
 	((int*)table)[SECTOR_SIZE/sizeof(int)-2] = crc32(table, SECTOR_SIZE-2*sizeof(int), CRC_POLY1);
