@@ -187,9 +187,10 @@ int has_partitions_beta(dev_t dev)
       major == MAJOR_IDE5 || 
 #endif
       major == MAJOR_IDE6 ||
+      major == MAJOR_MDP  ||  major == MAJOR_EMD  ||
       (major >= MAJOR_IDE7 && major <= MAJOR_IDE10) ||
       major == MAJOR_XT || major == MAJOR_ESDI || major == MAJOR_ACORN
-      	) return 0xFFC0;	/* 6 bit partition mask */
+      	) return 0xFFFFFFC0;	/* 6 bit partition mask */
       
     if (
       major == MAJOR_SD || (major >= MAJOR_SD2 && major <= MAJOR_SD8) ||
@@ -200,23 +201,23 @@ int has_partitions_beta(dev_t dev)
       (major >= MAJOR_CISS && major <= MAJOR_CISS+7) ||
       major == MAJOR_FTL || major == MAJOR_NFTL || major == MAJOR_DOC ||
       (major >= MAJOR_SD9 && major <= MAJOR_SD16) ||
-      major == MAJOR_EMD
-        ) return 0xFFF0;	/* 4 bit partition mask */
+      major == MAJOR_DM
+        ) return 0xFFFFFFF0;	/* 4 bit partition mask */
 
     if ( major == MAJOR_SATA || major == MAJOR_SATA2
-        )  return 0xFFE0;	/* 5 bit partition mask */
-
+        )  return 0xFFFFFFE0;	/* 5 bit partition mask */
 
     if ( major == MAJOR_IBM_iSER ||
       (major >= MAJOR_DAC960 && major <= MAJOR_DAC960+7) ||
       (major >= MAJOR_DAC960_8 && major <= MAJOR_DAC960_8+7)
-        )  return 0xFFF8;	/* 3 bit partition mask */
+        )  return 0xFFFFFFF8;	/* 3 bit partition mask */
 
     return 0;
 }
 #endif
 
-unsigned char max_partno[256] = {
+static
+unsigned char max_partno[512] = {
 /*
   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
 */
@@ -237,15 +238,16 @@ unsigned char max_partno[256] = {
   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
 */
  15, 15, 15, 15, 15, 15, 15, 15,  7,  7,  7,  7,  7,  7,  7,  7,   /*  8x   */
-  0,  0,  0,  0,  0,  0,  0,  0,  0, 15,  0,  0,  0,  0,  0,  0,
+  0,  0,  0,  0,  0,  0,  0,  0,  0, 63,  0,  0,  0,  0,  0,  0,
  31, 31,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
 
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   /*  Cx   */
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
-/*
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 15, 63,  0	   /*  /dev/md_dXX at 254  */
+								   /* and /dev/dm-0 at 253 */
+/*								   
   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
 */
 };
@@ -253,13 +255,13 @@ unsigned char max_partno[256] = {
 int has_partitions(dev_t dev)
 {
    int major = MAJOR(dev);
-   int ret;
+   int ret=0;
    
-   if (
-/***   	major < sizeof(max_partno) &&	***/
-   	(ret=max_partno[major])
-   					)  ret ^= 0xFFFF;
-   else ret = 0;
+   if (major >= nelem(max_partno)) {
+	warn("Major Device (%d) > %d",
+		major, nelem(max_partno)-1);
+   }
+   else if ( (ret=max_partno[major]) )  ret ^= 0xFFFFFFFF;
    
    return ret;
 }
@@ -272,26 +274,38 @@ void do_disk(void)
     char *disk,*bios,*sectors,*heads,*cylinders,*maxpart;
     int major;
 
-    entry = alloc_t(DT_ENTRY);
     disk = cfg_get_strg(cf_options,"disk");
-    if (stat(disk,&st) < 0) die("stat %s: %s",disk,strerror(errno));
+    cfg_init(cf_disk);
+    (void) cfg_parse(cf_disk);
+    if (stat(disk,&st) < 0) {
+        if (cfg_get_flag(cf_disk,"inaccessible")) {
+            cfg_unset(cf_options,"disk");
+            return;
+        }
+        die("do_disk: stat %s: %s",disk,strerror(errno));
+    }
     if (!S_ISBLK(st.st_mode) || 
     	(has_partitions(st.st_rdev) && (MINOR(st.st_rdev) & P_MASK(st.st_rdev))))
 		die(" '%s' is not a whole disk device",disk);
+
+    entry = alloc_t(DT_ENTRY);
     entry->device = st.st_rdev;
     major = MAJOR(st.st_rdev);
-    cfg_init(cf_disk);
-    (void) cfg_parse(cf_disk);
     bios = cfg_get_strg(cf_disk,"bios");
     sectors = cfg_get_strg(cf_disk,"sectors");
     heads = cfg_get_strg(cf_disk,"heads");
     cylinders = cfg_get_strg(cf_disk,"cylinders");
     maxpart = cfg_get_strg(cf_disk,"max-partitions");
     if (maxpart) {
+      if (major<nelem(max_partno)) {
 	int i = to_number(maxpart);
 	if (max_partno[major] && max_partno[major]!=i) die("Cannot alter 'max-partitions' for known disk  %s", disk);
 	max_partno[major] = i;
 	if (i!=7 && i!=15 && i!=31 && i!=63) die("disk=%s:  illegal value for max-partitions(%d)", disk, i);
+      }
+      else {
+        die("Implementation restriction: max-partitions on major device > %d", nelem(max_partno)-1);
+      }
     }
     entry->bios = bios ? to_number(bios) : -1;
     if (!sectors && !heads) entry->sectors = entry->heads = -1;
@@ -393,36 +407,18 @@ static int last_dev(int major,int increment)
  * only used to count IDE drives anyway, we try now only the first two devices
  * and forget about scan_last_dev.
  */
-#if 0
-    DEVICE dev;
-    int minor;
-
-    for (minor = 0; dev_open(&dev,(major << 8) | minor,O_BYPASS); minor += increment)
-	if (exists(dev.name)) dev_close(&dev);
-        else {
-	    dev_close(&dev);
-	    return minor/increment;
-	}
-    return scan_last_dev(NULL,DEV_DIR,major,increment);
-#else
     DEVICE dev;
     int devs;
 
-#if 0
-    for (devs = 0; devs < 2 && dev_open(&dev,(major << 8) | (increment*devs),
-      O_BYPASS); devs++)
-#else
     for (devs = 0;
 	devs < 2 && dev_open(&dev,MKDEV(major,increment*devs),O_BYPASS);
 	    devs++)
-#endif
 	if (exists(dev.name)) dev_close(&dev);
         else {
 	    dev_close(&dev);
 	    break;
 	}
     return devs;
-#endif
 }
 
 
@@ -474,7 +470,7 @@ void evms_bmap(struct evms_get_bmap_t *ebm)
         char evms_blk[] = "/dev/evms/block_device";
         struct evms_version_t evms_ver;
 
-        // Open the EVMS device
+        /* Open the EVMS device */
         if (evms_fd != -1)
             close(evms_fd);
 
@@ -482,11 +478,11 @@ void evms_bmap(struct evms_get_bmap_t *ebm)
         if (evms_fd < 0)
             die("Can't open EVMS block device %s.\n", evms_blk);
 
-        // Get EVMS ioctl version number.
+        /* Get EVMS ioctl version number. */
         if (ioctl(evms_fd, EVMS_GET_IOCTL_VERSION, &evms_ver) < 0)
             die("EVMS_GET_IOCTL_VERSION failed on %s.\n", evms_blk);
 
-        // Check that the ioctl version is >= 7.1.0
+        /* Check that the ioctl version is >= 7.1.0 */
         if (evms_ver.major < 7 ||
             (evms_ver.major == 7 && evms_ver.minor < 1))
             die("EVMS ioctl version %d.%d.%d does not support booting.\n",
@@ -597,6 +593,20 @@ void geo_query_dev(GEOMETRY *geo,int device,int all)
 	    geo->sectors = hdprm.sectors;
 	    geo->start = hdprm.start;
 	    break;
+	MASK31:
+	    geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 5);
+	    if (!get_all) break;
+	    if (ioctl(fd,HDIO_GETGEO,&hdprm) < 0)
+		die("geo_query_dev HDIO_GETGEO (dev 0x%04x): %s",device,
+		  strerror(errno));
+	    if (all && !hdprm.sectors)
+		die("HDIO_REQ not supported for your Disk controller. Please "
+		  "use a DISK section");
+	    geo->heads = hdprm.heads;
+	    geo->cylinders = hdprm.cylinders;
+	    geo->sectors = hdprm.sectors;
+	    geo->start = hdprm.start;
+	    break;
 	case MAJOR_DAC960:
 	case MAJOR_DAC960+1:
 	case MAJOR_DAC960+2:
@@ -670,10 +680,12 @@ void geo_query_dev(GEOMETRY *geo,int device,int all)
 	default:
 	    if (max_partno[major] && major==MAJOR_LOOP) break;
 	    if (max_partno[major] == 63)  goto MASK63;
+	    if (max_partno[major] == 31)  goto MASK31;
 	    if (max_partno[major] == 15)  goto MASK15;
 	    if (max_partno[major] == 7)   goto MASK7;
 	    
-	    if (MAJOR(device)>=120 && MAJOR(device)<=127)
+	    if ((MAJOR(device)>=120 && MAJOR(device)<=127)  ||
+	        (MAJOR(device)>=240 && MAJOR(device)<=254) )
 		die("Linux experimental device 0x04x needs to be defined.\n"
 		    "Check 'man lilo.conf' under 'disk=' and 'max-partitions='", device);
 	    else die("Sorry, don't know how to handle device 0x%04x",device);
@@ -946,9 +958,9 @@ int pass;
 	
 	if (all) bios_device(geo, device);
 	
-	if (!nowarn && (geo->device & 0x7f) >= bios_max_devs() &&
+	if ((geo->device & 0x7f) >= bios_max_devs() &&
 	  user_device == -1 && (!walk || walk->bios == -1))
-	    fprintf(errstd,"Warning: BIOS drive 0x%02x may not be accessible\n",
+	    warn("BIOS drive 0x%02x may not be accessible",
 	      geo->device);
     }
     if (walk) {
@@ -972,18 +984,21 @@ int pass;
     if (geo->heads > BIOS_MAX_HEADS)
 	die("Device 0x%04X: Maximum number of heads is %d, not %d\n",device,
 	  BIOS_MAX_HEADS,geo->heads);
-    if (geo->heads == BIOS_MAX_HEADS && !nowarn)
-	fprintf(errstd, "Warning: Maximum number of heads = %d (as specified)\n"
-			"   exceeds standard BIOS maximum of 255.\n", geo->heads);
+    if (geo->heads == BIOS_MAX_HEADS)
+	warn("Maximum number of heads = %d (as specified)\n"
+			"   exceeds standard BIOS maximum of 255.", geo->heads);
     if (geo->sectors > BIOS_MAX_SECS)
 	die("Device 0x%04X: Maximum number of sectors is %d, not %d\n",
 	  device,BIOS_MAX_SECS,geo->sectors);
+    if (geo->sectors == BIOS_MAX_SECS)
+        warn("Maximum number of heads = %d (as specified)\n"
+                        "   exceeds standard BIOS maximum of 63.", geo->sectors);
     if (!lba32 &&
-      (geo->start+geo->sectors-1)/geo->heads/geo->sectors >= BIOS_MAX_CYLS &&
-      !nowarn) {
-	fprintf(errstd,"Warning: device 0x%04x exceeds %d cylinder limit.\n"
-        "   Use of the 'lba32' option may help on newer (EDD BIOS) systems.\n",
-	  device,BIOS_MAX_CYLS);
+      (geo->start+geo->sectors-1)/geo->heads/geo->sectors >= BIOS_MAX_CYLS
+    ) {
+	warn("device 0x%04x exceeds %d cylinder limit.\n"
+        "   Use of the 'lba32' option may help on newer (EDD-BIOS) systems.",
+	  device, BIOS_MAX_CYLS);
     }
     if (verbose >= 3) {
 	printf("Device 0x%04x: BIOS drive 0x%02x, %d heads, %d cylinders,\n",
@@ -1010,9 +1025,9 @@ int geo_open(GEOMETRY *geo,char *name,int flags)
 
     if ((here = strrchr(name,':')) == NULL) user_dev = -1;
     else {
-	fprintf(errstd,":BIOS syntax is no longer supported. Please use a "
-	  "DISK section\n");
 	*here++ = 0;
+        warn("%s:BIOS syntax is no longer supported.\n    Please use a "
+	  "DISK section.", name);
 	user_dev = to_number(here);
     }
     if ((geo->fd = open(name,flags)) < 0)
@@ -1039,7 +1054,7 @@ int geo_open(GEOMETRY *geo,char *name,int flags)
     if (!geo->file) geo->spb = 2;
     else {
 	if (ioctl(geo->fd,FIGETBSZ,&block_size) < 0) {
-	    fprintf(errstd,"FIGETBSZ %s: %s\n",name,strerror(errno));
+	    warn("FIGETBSZ %s: %s",name,strerror(errno));
 	    geo->spb = 2;
 	}
 	else {
@@ -1097,7 +1112,7 @@ void geo_close(GEOMETRY *geo)
 int geo_comp_addr(GEOMETRY *geo,int offset,SECTOR_ADDR *addr)
 {
     int block,sector;
-    static int warnings = 0;
+    static int linear_warnings = 0;
 
 #if BETA_TEST
     if (verbose>=6)
@@ -1199,9 +1214,8 @@ int geo_comp_addr(GEOMETRY *geo,int offset,SECTOR_ADDR *addr)
 	    if (geo->sectors>0 && geo->heads>0) {
 	    	cyl /= geo->sectors;
 	    	cyl /= geo->heads;
-	    	if (cyl >= BIOS_MAX_CYLS && !nowarn && warnings++ < 8) {
-		    fprintf(errstd,
-			"Warning:  LINEAR may generate cylinder# above 1023 at boot-time.\n");
+	    	if (cyl >= BIOS_MAX_CYLS && linear_warnings++ < 8) {
+		    warn("LINEAR may generate cylinder# above 1023 at boot-time.");
 	    	}
 	    }
             if (sector/(63*255) >= BIOS_MAX_CYLS)
