@@ -1,7 +1,7 @@
 /* lilo.c  -  LILO command-line parameter processing */
 /*
 Copyright 1992-1998 Werner Almesberger.
-Copyright 1999-2005 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -20,7 +20,6 @@ source directory.
 #include <errno.h>
 #include <time.h>
 #include <limits.h>
-#include <sys/stat.h>
 #include <dirent.h>
 /*#include <asm/page.h>*/
 
@@ -117,6 +116,7 @@ static char buffer[SECTOR_SIZE];
     int tsecs;
     int tlinear, tlba32;
     unsigned short flags;
+    time_t Time;
 
 #if !__MSDOS__
     fd = geo_open(&geo,map_file,O_RDONLY);
@@ -160,7 +160,12 @@ static char buffer[SECTOR_SIZE];
 		die("Cannot recognize boot sector.");
 	}
 #endif
+#if 1
+        Time = boot.par_1.map_stamp;
+	printf("Installed:  %s\n", ctime(&Time));
+#else
 	printf("Installed:  %s\n", ctime((time_t*)&boot.par_1.map_stamp));
+#endif
 	printf("Global settings:\n");
 	tsecs = (param2.delay*2197+3999)/4000;
 	printf("  Delay before booting: %d.%d seconds\n",tsecs/10,tsecs % 10);
@@ -214,10 +219,14 @@ static char buffer[SECTOR_SIZE];
 #if __MSDOS__
 	    printf("%s\n", name
 #else /* !__MSDOS__ */
-	    fprintf(stdout,"%s%-" S(MAX_IMAGE_NAME) "s %s%s",verbose > 0 ? "  " : "",name,
+	    printf("%s%-" S(MAX_IMAGE_NAME) "s %s%s%s",verbose > 0 ? "  " : "",name,
 	      image ? "" : "*",
 #ifdef LCF_VIRTUAL
 	      descrs.d.descr[image].flags & FLAG_VMDEFAULT ? "@" :
+#endif
+	      "",
+#ifdef LCF_NOKEYBOARD
+	      descrs.d.descr[image].flags & FLAG_NOKBDEFAULT ? "&" :
 #endif
 	      ""
 #endif /* !__MSDOS__ */
@@ -253,6 +262,10 @@ static char buffer[SECTOR_SIZE];
 		    printf("    Virtual Boot is disabled\n");
 		if (flags & FLAG_VMWARN)
 		    printf("    Warn on Virtual boot\n");
+#endif		
+#ifdef LCF_NOKEYBOARD
+		if (flags & FLAG_NOKBDISABLE)
+		    printf("    NoKeyboard Boot is disabled\n");
 #endif		
 		if ( !(flags & FLAG_PASSWORD) )
 		    printf("    No password\n");
@@ -486,6 +499,9 @@ fprintf(errstd,"argc=%d, *argv=%s, ch=%c param=%s\n", argc, *argv, ch, param);
 	    case 'g':
 		geometric |= AD_GEOMETRIC;
 		break;
+	    case 'H':
+		force_raid=1;
+		break;
 	    case 'i':
 		cfg_set(cf_options,"install",param,NULL);
 		break;
@@ -698,7 +714,7 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 
 #if BETA_TEST
 {
-    int i, k=0;
+    int i, j, k=0;
 extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 
 
@@ -709,7 +725,22 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 	    k ++;
 	}
     }
-    printf("k = %d (should be 0)\n\n", k);
+    printf("%sk = %d (should be 0)\n\n", k?"ERROR: ":"", k);
+    for (i=0, j=0; i<40960; i+=17) {
+        dev_t dev;
+        int major, minor;
+        for (k=0; k<81920; k+=19) {
+            dev = MKDEV(i,k);
+            major = MAJOR(dev);
+            minor = MINOR(dev);
+            if (major != i  ||  minor != k) {
+                printf("ERROR: (%d,%d) -> (%d,%d)\n",
+                    i, k, major, minor);
+                j++;
+            }
+        }
+    }
+    if (j==0) printf("MKDEV check passed\n");
 }
 #endif
 
@@ -731,6 +762,7 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 			LINUX_VERSION_CODE>>16,
 			LINUX_VERSION_CODE>>8 & 255,
 			LINUX_VERSION_CODE & 255);
+		printf("Maximum Major Device = %d\n", MAJOR(~0UL));
 #endif /* !__MSDOS__ */
 		printf("\nMAX_IMAGES = %d\t\tc=%d, s=%d, i=%d, "
 				"l=%d, ll=%d, f=%d, d=%d, ld=%d\n\n",
@@ -767,8 +799,8 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     if (new_root && *new_root) {
 	pp_fd = fopen(PARTITIONS, "r");
 	if (chroot(new_root) < 0) die("chroot %s: %s",new_root,strerror(errno));
-	if (chdir("/dev") < 0  &&  nowarn < 1 )
-	    fprintf(errstd, "Warning: root at %s has no /dev directory\n", new_root);
+	if (chdir("/dev") < 0)
+	        warn("root at %s has no /dev directory", new_root);
 	if (chdir("/") < 0) die("chdir /: %s",strerror(errno));
     }
     if (atexit(temp_remove)) die("atexit() failed");
@@ -783,7 +815,7 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 	    return 0;
 	}
 	printf(", Copyright (C) 1992-1998 Werner Almesberger\n"
-	       "Development beyond version 21 Copyright (C) 1999-2005 John Coffman\n"
+	       "Development beyond version 21 Copyright (C) 1999-2006 John Coffman\n"
 	       );
         if (verbose>0) {
 #if VERSION_MINOR>=50 && !__MSDOS__
@@ -820,29 +852,26 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     
 #if !__MSDOS__
     temp = cfg_get_flag(cf_options,"nowarn");
-#endif /* !__MSDOS__ */
     if (nowarn < 0) nowarn = 0;
     else nowarn = temp;
+/* All warnings appear if very verbose modes used */
+    if (verbose>=3) nowarn = 0;
+#endif /* !__MSDOS__ */
 
     if (verbose>=6) printf("main: cfg_parse returns %d\n", more);
 
 #if !__MSDOS__
     if (tell_param && !tell_early) probe_tell(tell_param);
 
-    if (!nowarn) {
-	if (fstat(fd,&st) < 0) {
-	    fprintf(errstd,"fstat %s: %s\n",config_file,strerror(errno));
-	    exit(1);
-	}
-	if (S_ISREG(st.st_mode)) {
-	    if (st.st_uid)
-		fprintf(errstd,"Warning: %s should be owned by root\n",
-		  config_file);
-	    else if (st.st_mode & (S_IWGRP | S_IWOTH))
-		    fprintf(errstd,"Warning: %s should be writable only for "
-		      "root\n",config_file);
-	    config_read = !!(st.st_mode & (S_IRGRP | S_IROTH));
-	}
+    if (fstat(fd,&st) < 0)
+	    die("fstat %s: %s", config_file, strerror(errno) );
+
+    if (S_ISREG(st.st_mode)) {
+        if (st.st_uid)
+            warn("%s should be owned by root", config_file);
+        else if (st.st_mode & (S_IWGRP | S_IWOTH))
+	    warn("%s should be writable only for root", config_file);
+        config_read = !!(st.st_mode & (S_IRGRP | S_IROTH));
     }
     preload_dev_cache();
     
@@ -859,23 +888,23 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     if (geom+linear+lba32 > 1)
 	die("May specify only one of GEOMETRIC, LINEAR or LBA32");
     if (geometric) {
-	if (!nowarn && (geom+linear+lba32 > 0))  
-	    fprintf(errstd,"Ignoring entry '%s'\n", geom ? "geometric" :
-	    	linear ? "linear" : "lba32");
+	if (geom+linear+lba32 > 0)  
+	    warn("Ignoring entry '%s'", geom ? "geometric" :
+	    	      linear ? "linear" : "lba32");
 	geom = linear = lba32 = 0;
 	if (geometric==AD_LBA32) lba32 = 1;
 	else if (geometric==AD_LINEAR) linear = 1;
 	else if (geometric==AD_GEOMETRIC) geom = 1;
     }    
     if (geom+linear+lba32 == 0) {
-	if (!nowarn && verbose) fprintf(errstd,"Warning: LBA32 addressing assumed\n");
+	warn("LBA32 addressing assumed");
 	lba32 = 1;
     }
-    if (linear && !nowarn) fprintf(errstd,
-    	"\nWarning:  LINEAR is deprecated in favor of LBA32:  LINEAR specifies 24-bit\n"
+    if (linear) warn(
+    	"LINEAR is deprecated in favor of LBA32:  LINEAR specifies 24-bit\n"
 	"  disk addresses below the 1024 cylinder limit; LBA32 specifies 32-bit disk\n"
-	"  addresses not subject to cylinder limits on systems with EDD bios extensions;\n"
-	"  use LINEAR only if you are aware of its limitations.\n\n");
+	"  addresses not subject to cylinder limits on systems with EDD-BIOS extensions;\n"
+	"  use LINEAR only if you are aware of its limitations.");
     
     if (identify) identify_image(identify,ident_opt);
 
@@ -897,9 +926,9 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     }
     if (bios_passes_dl == DL_NOT_SET) 	check_bios();	/* in probe.c */
 
-    if (!nowarn && compact && (linear || lba32) && verbose>=2)
-	fprintf(errstd,"Warning: COMPACT may conflict with %s on some "
-		"systems\n", lba32 ? "LBA32" : "LINEAR");
+    if (compact && (linear || lba32) && verbose>=4)
+	warn("COMPACT may conflict with %s on some "
+		"systems", lba32 ? "LBA32" : "LINEAR");
 
     geo_init(cfg_get_strg(cf_options,"disktab"));
 #endif /* !__MSDOS__ */
@@ -940,7 +969,13 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 	temp = bsect_number();
 	if (temp==0) die("No images have been defined.");
 	else if (temp<0) die("Default image doesn't exist.");
-	
+
+#ifdef LCF_VIRTUAL
+	check_vmdefault();
+#endif
+#ifdef LCF_NOKEYBOARD
+	check_nokbdefault();
+#endif
 	check_fallback();
 	check_unattended();
 	
@@ -965,6 +1000,12 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 	      "altered.\n");
 	}
 	if (verbose>=4) dump_serial_nos();
+	if (warnings) {
+	    if (warnings>1)
+	        printf("%d warnings were ", warnings);
+            else printf("One warning was ");
+            printf("%sed.\n", nowarn ? "suppress" : "issu");
+	}
 #else  /* __MSDOS__ */
 	die("No option switches specified:  -q, -R, or -V");
 #endif /* __MSDOS__ */
