@@ -1,7 +1,7 @@
 /* lilo.h  -  LILO constants
 
 Copyright 1992-1998 Werner Almesberger.
-Copyright 1999-2004 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -28,6 +28,9 @@ source directory.
 #if defined(LCF_READONLY) && defined(LCF_REWRITE_TABLE)
 #error "Incompatible Makefile options: READONLY and REWRITE_TABLE"
 #endif
+#ifdef LCF_LARGE_EBDA
+#error "Configuration option '-DLARGE_EBDA' is deprecated"
+#endif
 
 /*
  * Starting with 2.1.something, Linux kernels put VGA constants and segment
@@ -90,8 +93,10 @@ source directory.
 
 /* the known major device numbers */
 #define MAJMIN_RAM	0x101 /* RAM disk */
+#define MAJOR_MDP_SORT  1 /* Convert MAJOR_MDP to this for sorting */
+#define MAJOR_EMD_SORT  2 /* Convert MAJOR_EMD to this for sorting */
+#define MAJOR_HPT370_SORT  2 /* Convert MAJOR_HPT320 to this for sorting */
 #define MAJOR_FD	2 /* floppy disks */
-#define MAJOR_HPT370_SORT  MAJOR_FD	/* 114->2 for sorting HPT370 controller */
 #define MAJOR_HD	3 /* IDE-type hard disks */
 #define MAJOR_LOOP	7 /* Loopback devices 0-15 */
 #define MAJOR_SD	8 /* SCSI disks 0-15 */
@@ -142,6 +147,8 @@ source directory.
 #define MAJOR_EMD	153 /* Enhanced multi-disk RAID sets */
 #define MAJOR_SATA	160 /* Carmel SATA Disk on first 8-port controller */
 #define MAJOR_SATA2	161 /* Carmel SATA Disk on 2nd 8-port controller */
+#define MAJOR_DM	253 /* linux/drivers/md/dm.c RAID set [experimental?] */
+#define MAJOR_MDP	254 /* Enhanced multi-disk RAID sets [experimental?] */
 
 #define MAX_TOKEN	1023 /* max device Token length */
 #define MAX_IMAGE_NAME	15 /* maximum name length (w/o terminating NUL) */
@@ -164,12 +171,12 @@ source directory.
 #define MAX_DESCR_SECS 2	/* maximum # of descriptor sectors */
 #endif
 #if MAX_DESCR_SECS > MAX_DESCR_SECTORS
-# error "Maximum DSECS=x exceeded."
+#error "Maximum DSECS=x exceeded."
 #endif
 #define MAX_IMAGES      ((int)((SECTOR_SIZE*MAX_DESCR_SECS-sizeof(INT4)-1)/sizeof(IMAGE_DESCR)))
 			  /* maximum number of images */
 
-#define COMMAND_LINE_SIZE	256	/* CL_LENGTH */
+#define COMMAND_LINE_SIZE	512	/* CL_LENGTH */
 #define SECTOR_SIZE	512 /* disk sector size */
 #ifndef BLOCK_SIZE /* may also come from linux/fs.h */
 #define BLOCK_SIZE	1024 /* disk block size */
@@ -182,7 +189,7 @@ source directory.
 #define PART_ACT_ENT_OFF   0	/* offset in entry for active flag */
 #define PART_TYPE_ENT_OFF  4	/* offset in entry for partition type */
 
-#define P_MASK(x) ((~has_partitions(x))&0xFFFF)
+#define P_MASK(x) ((~has_partitions(x))&0xFFFFFFFF)
 #define D_MASK(x) (has_partitions(x))
 #define PART_MAX PARTITION_ENTRIES /* biggest primary partition number */
 #define PART_MAX_MAX	63	/* max. partition number (on IDE disks) */
@@ -229,9 +236,9 @@ source directory.
 #define PART_INVALID	  98	/* invalid partition type */
 
 #define HIDDEN_OFF	0x10	/* type offset to hide partition (OS/2 BM) */
-#define PART_HDOS12	PART_DOS12+HIDDEN_OFF
-#define PART_HDOS16_SMALL PART_DOS16_SMALL+HIDDEN_OFF
-#define PART_HDOS16_BIG	PART_DOS16_BIG+HIDDEN_OFF
+#define PART_HDOS12	(PART_DOS12+HIDDEN_OFF)
+#define PART_HDOS16_SMALL (PART_DOS16_SMALL+HIDDEN_OFF)
+#define PART_HDOS16_BIG	(PART_DOS16_BIG+HIDDEN_OFF)
 
 #define STAGE_FIRST	1	/* first stage loader code */
 #define STAGE_SECOND	2	/* second stage loader code */
@@ -282,7 +289,7 @@ source directory.
 #define BIOS_MAX_HEADS  256	/* 8 bits head number; really 255 */
 			/* but must account for oddball BIOS's that allow 256 */
 #define BIOS_MAX_CYLS   1024	/* 10 bits cylinder number */
-#define BIOS_MAX_SECS   63	/* 6 bits sector number (no, it's not 64) */
+#define BIOS_MAX_SECS   64	/* 6 bits sector number (really 63) */
 
 /* these are the boot record flags in the "prompt" variable */
 #define FLAG_PROMPT	1	/* always issue boot: prompt */
@@ -302,8 +309,10 @@ source directory.
 #endif
 
 /* these are the second-stage specific flags */
-#define FLAG2_EL_TORITO	 2	/* El Torito format bootable CD */
+#define FLAG2_EL_TORITO  2	/* El Torito format bootable CD */
 #define FLAG2_UNATTENDED 4	/* Unattended booting option */
+#define FLAG2_VIRTUAL	 8	/* vmdefault, vmdisable, vmwarn used */
+#define FLAG2_NOKBD	16	/* nokbdefault, nokbdisable used */
 
 /* these are the descriptor flags */
 #define FLAG_VGA	1	/* override VGA mode */
@@ -322,6 +331,10 @@ source directory.
 #endif
 #define FLAG_SINGLE	4096	/* single key activation */
 #define FLAG_RETAIN	0x2000	/* retain BMP screen on boot */
+#ifdef LCF_NOKEYBOARD
+#define FLAG_NOKBDEFAULT 0x4000	/* this is the default with no keyboard */
+#define FLAG_NOKBDISABLE 0x8000	/* unable to boot if no keyboard */
+#endif
 
 #define VGA_NOCOVR	0x8000	/* VGA setting not overridden on command line */
 
