@@ -1,6 +1,6 @@
 /* probe.c -- BIOS probes */
 /*
-Copyright 1999-2004 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -190,7 +190,7 @@ int fetch(void)
         buf.s5.version > (short)(PROBE_VERSION)) return buf_valid=3;
     got = buf.s5.length;
     if (got > sizeof(buf.b) || got < sizeof(buf.s4)) return buf_valid=4;
-    if (*(int*)buf.s5.checksum != crc32((char*)&buf.s5 + 4, got-4, CRC_POLY1))
+    if (*(int*)buf.s5.checksum != crc32((unsigned char*)&buf.s5 + 4, got-4, CRC_POLY1))
     	return buf_valid=5;
 
     if (buf.s5.version == 4) {
@@ -229,7 +229,7 @@ int purge(void)
     if ((i=fetch())) return i;	/* return error from fetch */
     
     i = buf.s5.length;
-    *(int*)buf.s5.checksum = crc32((char*)&buf.s5 + 4, i-4, CRC_POLY1);
+    *(int*)buf.s5.checksum = crc32((unsigned char*)&buf.s5 + 4, i-4, CRC_POLY1);
     
     if ((fd=open("/dev/mem", O_WRONLY)) < 0) pdie("purge: can't open /dev/mem");
     if (lseek(fd, seek, SEEK_SET) != seek) pdie("purge: ");
@@ -259,7 +259,7 @@ static int notice(int needed)
 		"Booting your system with LILO version %s or later would provide the re-\n"
 		"quested information as part of the BIOS data check.  Please install a more\n"
 		"recent version of LILO on your hard disk, or create a bootable rescue floppy\n"
-		"with the 'mkrescue' command.\n\n",
+		"or rescue CD with the 'mkrescue' command.\n\n",
 		
 		needed==4 ? "22.0" :
 		needed==5 ? "22.5.1" :
@@ -297,6 +297,7 @@ static void do_help(void)
 /* diagnostic output */
 static void show_geom(char *who, int cyl, int head, int sect)
 {
+   if (nowarn) return;
    fprintf(errstd, "    %s: %d cylinders, %d heads, %d sectors\n", who, cyl, head, sect);
 }
 
@@ -309,6 +310,8 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
     int i;
     struct partition *pt_base;
     unsigned int total;
+    int sec_err = 0;
+    int hd_err = 0;
 
 #if 0
     if((i=fetch())) {
@@ -377,8 +380,9 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
    if (hd->fn08.flags&1 || hd->fn08.ah || hd->fn08.cx==0)
      return 1 + hd->fn08.ah;
    
-   i =
-   geom->n_sect = hd->fn08.cx & 0x3F;
+   if (!(i = hd->fn08.cx & 0x3F))  i = 64;	/* BIOS bug if 0 */
+   geom->n_sect = i;
+   
    i *=
    geom->n_head = ((hd->fn08.dx>>8)&0xFF)+1;
    i *=
@@ -387,6 +391,9 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
    geom->n_disks = hd->fn08.dx & 0xFF;
    geom->pt = NULL;
 
+   hd_err = (geom->n_head > 255);
+   sec_err = (geom->n_sect > 63);
+   
    if (drive < 4)  return 0;
    
    pt_base = NULL;
@@ -446,9 +453,9 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	      if ((dp->info) & EDD_PARAM_GEOM_VALID) {
 	         if ((geom->n_sect != dp->sectors || geom->n_head != dp->heads) &&
 				((verbose>0 && !lba32) || verbose>=4) && 
-				!nowarn && !(warned[drive-0x80]&1) ) {
-	                fprintf(errstd,"Warning: Int 0x13 function 8 and function 0x48 return different\n"
-                	               "head/sector geometries for BIOS drive 0x%02X\n", drive);
+				!(warned[drive-0x80]&1) ) {
+	                warn("Int 0x13 function 8 and function 0x48 return different\n"
+                	               "head/sector geometries for BIOS drive 0x%02X", drive);
         	        show_geom("fn 08", geom->n_cyl, geom->n_head, geom->n_sect);
 	                show_geom("fn 48", dp->cylinders, dp->heads, dp->sectors);
 			warned[drive-0x80] |= 1;
@@ -469,7 +476,20 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	      if (dp->total_sectors > geom->n_total_blocks)
 			geom->n_total_blocks = dp->total_sectors;
 	}
-   }
+    }
+    if (!(warned[drive-0x80]&4)) {
+       if (hd_err) warn("LILO is compensating for a BIOS bug: (drive 0x%02X) heads > 255",
+                           drive);
+       if (sec_err) {
+           warn("LILO will try to compensate for a BIOS bug: (drive 0x%02X) sectors > 63",
+                           drive);
+           if ((geom->EDD_flags & EDD_PACKET) && !lba32)
+               die("LBA32 addressing should be used, not %s", linear ? "LINEAR" : "GEOMETRIC");
+           if (!(geom->EDD_flags & EDD_PACKET)   &&   !(lba32 | linear) )
+               warn("Drive 0x%02X may not be usable at boot-time.", drive);
+       }
+       warned[drive-0x80] |= 4;
+    }
    
    return 0;
 }
@@ -507,7 +527,7 @@ static void do_ebda(void)
     m <<= 10;
     m -= 0x200;
     n = (select_loader()->size + SECTOR_SIZE - 1) / SECTOR_SIZE;
-    n = m - (n+4+MAX_DESCR_SECS)*SECTOR_SIZE;
+    n = m - (n+4+MAX_DESCR_SECS+(COMMAND_LINE_SIZE>256))*SECTOR_SIZE;
     init = (n - (MAX_SETUPSECS+1)*SECTOR_SIZE)>>4;
     if (init > DEF_INITSEG) init = DEF_INITSEG;
     printf("\n");
@@ -1008,8 +1028,8 @@ int bios_device(GEOMETRY *geo, int device)
 	       (geo->heads && geo->heads!=bdata.n_head)  )  {
 	    unsigned int nblocks = geo->cylinders * geo->heads * geo->sectors;
 
-	    if (verbose>0 && !nowarn && !(warned[bios1-0x80]&2) ) {
-		fprintf(errstd,"Warning: Kernel & BIOS return differing head/sector geometries for device 0x%02X\n", bios1);
+	    if (!(lba32 | linear) && !(warned[bios1-0x80]&2) ) {
+		warn("Kernel & BIOS return differing head/sector geometries for device 0x%02X", bios1);
 		show_geom("Kernel", geo->cylinders, geo->heads, geo->sectors);
 		show_geom("  BIOS", bdata.n_cyl, bdata.n_head, bdata.n_sect);
 		warned[bios1-0x80] |= 2;
